;; Generated package description from consult-recoll.el  -*- no-byte-compile: t -*-
(define-package "consult-recoll" "0.7" "Recoll queries using consult" '((emacs "26.1") (consult "0.18")) :commit "714810486a13bb6adba5be0eaaca7eabe3298927" :authors '(("Jose A Ortega Ruiz" . "jao@gnu.org")) :maintainer '("Jose A Ortega Ruiz" . "jao@gnu.org") :keywords '("docs" "convenience") :url "https://codeberg.org/jao/consult-recoll")
